import glob
import random, os, json
import numpy as np

import tensorflow as tf
from transformers import AutoTokenizer
import logging

import tensorflow_addons as tfa
from tensorflow import keras
from keras import backend as K

class LRSchudlerCallback(keras.callbacks.Callback):
    def __init__(self, args, logger):
        super(LRSchudlerCallback, self).__init__()
        self.warmup_ratio = args.warmup_ratio
        self.scheduler = args.scheduler
        self.logger = logger

    def on_train_begin(self, logs=None):
        self.steps_per_epoch = self.params["steps"]
        self.epochs = self.params["epochs"]
        self.global_step = 0
        self.logger.info(f"using learning rate scheduler {self.scheduler}")
        if not self.scheduler.startswith("constant"):
            self.total_steps = self.steps_per_epoch * self.epochs
            self.warmup_steps = int(self.total_steps * self.warmup_ratio)
            self.logger.info(
                f"total_steps: {self.total_steps}, steps_per_epoch: {self.steps_per_epoch}, epochs: {self.epochs}, warmup_steps:{self.warmup_steps}")
            if not hasattr(self.model.optimizer, "lr"):
                raise ValueError('Optimizer must have a "lr" attribute.')
            self.logger.info(f"lr of optimizer to reach through warmup: {K.eval(self.model.optimizer.lr)}")

            self.lr_to_reach = K.eval(self.model.optimizer.lr)
            K.set_value(self.model.optimizer.learning_rate, 0.00)
            self.logger.info(f"now set it to zero for warmup: {K.eval(self.model.optimizer.lr)}")


    def on_train_batch_end(self, batch, logs=None):
        if self.global_step <= self.warmup_steps:
            if self.scheduler == "warmuplinear" or self.scheduler == "warmupconstant":
                inc = self.lr_to_reach / self.warmup_steps
                K.set_value(self.model.optimizer.learning_rate, K.eval(self.model.optimizer.lr) + inc)
        else:
            if self.scheduler == "warmuplinear" or self.scheduler == "constantlinear":
                dec = self.lr_to_reach / (self.total_steps - self.warmup_steps)
                K.set_value(self.model.optimizer.learning_rate, K.eval(self.model.optimizer.lr) - dec)

        self.global_step += 1

    def on_test_batch_end(self, batch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        self.logger.info(f"at epoch={epoch}, the learning_rate is {K.eval(self.model.optimizer.lr)}")

    def on_train_end(self, logs=None):
        self.logger.info("testing")


def get_callbacks(args, inputs, logger, eval_getter):
    tqdm_callback = tfa.callbacks.TQDMProgressBar(metrics_format="{name}: {value:0.8f}",
                                                  epoch_bar_format="{n_fmt}/{total_fmt}{bar} ETA: {remaining}s - {desc}, {rate_fmt}{postfix}", )
    lr_scheduler_callback = LRSchudlerCallback(args, logger)
    if args.do_eval == True:
        eval_callback = eval_getter(inputs["x_eval"], inputs["y_eval"], args)
        # return [tqdm_callback,eval_callback]
        return [tqdm_callback, eval_callback, lr_scheduler_callback]
    else:
        return [tqdm_callback, lr_scheduler_callback]


def create_model(args, logger, model_getter,tokenizer=None,from_pretrained=True,save_args=True):
    if args.use_tpu:
        # Create distribution strategy
        tpu = tf.distribute.cluster_resolver.TPUClusterResolver(tpu='grpc://' + args.tpu_address)
        tf.config.experimental_connect_to_cluster(tpu)
        tf.tpu.experimental.initialize_tpu_system(tpu)
        logger.info("All TPU devices: ")
        for each_device in tf.config.list_logical_devices('TPU'):
            logger.info(each_device)
        strategy = tf.distribute.TPUStrategy(tpu)
        # Create model
        with strategy.scope():
            model = model_getter(args,tokenizer=tokenizer,from_pretrained=from_pretrained)
    else:
        if args.use_gpu:
            # Create a MirroredStrategy.
            strategy = tf.distribute.MirroredStrategy()
            logger.info("Number of GPU devices: {}".format(strategy.num_replicas_in_sync))
            # Open a strategy scope.
            with strategy.scope():
                model = model_getter(args,tokenizer=tokenizer,from_pretrained=from_pretrained)
        else:
            raise ValueError("not available yet")
            # strategy = None
            # logger.info("Using CPU for training")
            # model = model_getter(args)

    logger.info(model.summary())
    # trainable_count = int(
    #     np.sum([K.count_params(p) for p in set(model.trainable_weights)]))
    # non_trainable_count = int(
    #     np.sum([K.count_params(p) for p in set(model.non_trainable_weights)]))
    # logger.info('Total params: {:,}'.format(trainable_count + non_trainable_count))
    # logger.info('Trainable params: {:,}'.format(trainable_count))
    # logger.info('Non-trainable params: {:,}'.format(non_trainable_count))
    # if strategy!=None:
    args.num_replicas_in_sync = strategy.num_replicas_in_sync
    if save_args:
        write_args(args.output_path, args)
    return model, strategy

def get_tokenizer(args):
    tokenizer = AutoTokenizer.from_pretrained(args.model_select)
    tokenizer.save_pretrained(args.output_path)
    return tokenizer

def add_filehandler_for_logger(output_path, logger, out_name="train"):
    logFormatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s')
    fileHandler = logging.FileHandler(os.path.join(output_path, f"{out_name}.log"), mode="a")
    fileHandler.setFormatter(logFormatter)
    logger.addHandler(fileHandler)

def set_seed(seed):
    tf.random.set_seed(
        seed
    )
    random.seed(seed)
    np.random.seed(seed)

def write_args(output_path, args):
    with open(os.path.join(output_path, "args.json"), "w") as f:
        f.write(json.dumps(args.__dict__, indent=2))

def get_existing_cks(output_path, best_ck=False):
    cks_path_already = glob.glob(os.path.join(output_path, "ck*.h5"))

    if best_ck:
        for ex in glob.glob(os.path.join(output_path, "best*.h5")):
            os.remove(ex)

    index2path = {int(os.path.basename(each_ck_path).split(".")[0].split("_")[-1]): each_ck_path for
                  each_ck_path in cks_path_already}
    sorted_indices = sorted(index2path)  # index here refers to the epoch number
    return sorted_indices, index2path

def load_torch_state_dict_from_h5_weights(model):
    import torch
    state_dict = {}
    for layer in model.layers:
        for resource_variable in layer.weights:
            key = resource_variable.name
            value = torch.tensor(resource_variable.numpy())
            state_dict[key] = value
    total_params_num = sum([element.numel() for element in state_dict.values()])
    print(f"the number of params: {total_params_num}")
    return state_dict

def save_transformer_locally(model_name="bert-base-uncased",save_path=".",is_tf=False):
    """save
    anyone you can find from here: https://huggingface.co/models
    """
    # to use AutoModel, need to install pytorch: pip3 install torch torchvision or pip install torch torchvision
    from transformers import AutoTokenizer, AutoModel, TFAutoModel
    if is_tf:
        model = TFAutoModel.from_pretrained(model_name)
    else:
        model = AutoModel.from_pretrained(model_name)

    # Load pretrained model/tokenizer
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    if not os.path.isdir(save_path):
        os.makedirs(save_path,exist_ok=True)
    model.save_pretrained(os.path.join(save_path,model_name))  # save model weights and config
    tokenizer.save_pretrained(os.path.join(save_path,model_name))  # save tokenizer config or/and vocab

def iid_denoise_text(original_text, span_length=3, corrupt_ratio=0.15, lang="zh_cn"):
    """
    This method is implemented for the pre-training objective of T5, as described in the T5 paper (https://arxiv.org/abs/1910.10683)
    this default params setup keeps the same as the original T5 paper on English, we generalize it to more languages such as Chinese
    :param original_text: it is a list of tokens
    :param span_length: 3 for by default as described in T5 paper
    :param corrupt_ratio: 15% by default as described in T5 paper
    :param lang: reserved param for future use
    :return:
    """
    source_text = []
    target_text = []
    # if lang == "en":
    #     corrupt_ratio = 0.15
    #     span_length = 3  # 3 as in T5 paper
    # make deterministic for reproducibility
    # random.seed(2020)
    replace_i = 0
    skip_count = span_length
    last_replace_pos = -span_length
    for pos in range(len(original_text)):
        if skip_count < span_length - 1:
            skip_count += 1
        else:
            if random.uniform(0, 1) < corrupt_ratio:
                extra_token = f"<extra_id_{replace_i}>"
                if pos != last_replace_pos + span_length:
                    target_text.append(extra_token)
                to_replace_span = original_text[pos: pos + span_length]
                target_text.extend(to_replace_span)
                source_text.append(extra_token)
                replace_i += 1
                skip_count = 0
                last_replace_pos = pos
            else:
                source_text.append(original_text[pos])
    if target_text == "" or target_text == []:
        target_text.append("<extra_id_0>")
    return original_text, source_text, target_text


if __name__ == '__main__':
    model_name_or_path = "t5-small"
    from transformers import TFT5ForConditionalGeneration
    model = TFT5ForConditionalGeneration.from_pretrained(model_name_or_path)
    state_dict = load_torch_state_dict_from_h5_weights(model)
    print(len(state_dict))


''''
this is a evaluation callback class for high-level Keras training (BERT-like models in this lib)
'''

import sys
from tensorflow import keras
import numpy as np
from sklearn.metrics import classification_report
import tensorflow as tf
import logging
import os
from ttt.utils import add_filehandler_for_logger, get_existing_cks
from tensorboardX import SummaryWriter

logging.basicConfig(
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)
logger = logging.getLogger(__name__)

class ClsEvaluator(keras.callbacks.Callback):
    def __init__(self, x_eval, y_eval, args):
        super(ClsEvaluator).__init__()
        self.x_eval = x_eval
        self.y_eval = y_eval
        self.eval_on = args.eval_on
        self.patience = args.patience
        self.log_steps = args.log_steps
        self.args = args
        self.use_tb = self.args.use_tb
        if self.use_tb:
            self._tb_writer = SummaryWriter(log_dir=os.path.join("runs", args.output_folder))

    def on_train_begin(self, logs=None):
        self.global_step = 0
        self.wait = 0
        self.best = np.Inf if self.eval_on == "loss" else -np.Inf

    def on_train_end(self, logs=None):
        if self.use_tb:
            self._tb_writer.close()

    def on_batch_end(self, batch, logs=None):
        self.global_step += 1
        if self.log_steps != -1 and self.global_step % self.log_steps == 0:
            if self.args.do_eval:
                self.evaluate(self.global_step, tag="global_step", logs=logs)

    def evaluate(self, steps, tag="epoch", logs=None):
        logger.info("\n")
        logger.info(f"*************evaluating at {tag}={steps}*************")
        eval_results = self.model.evaluate(self.x_eval, self.y_eval, batch_size=self.args.eval_batch_size)
        dev_loss, acc = eval_results[0], eval_results[1]
        pred_probs = self.model.predict(self.x_eval, batch_size=self.args.eval_batch_size)
        preds = tf.math.argmax(pred_probs, 1).numpy()
        # acc = accuracy_score(preds, self.y_eval)
        logger.info(f"{tag}={steps}, eval_report: {classification_report(self.y_eval, preds, digits=4)}")
        logger.info(f"{tag}={steps}, eval_acc: {acc}")
        logger.info(f"{tag}={steps}, eval_results: {eval_results}")
        logger.info(f"{tag}={steps}, train_logs: {logs}")

        if self.use_tb:
            if logs is not None:
                logger.info("logging metrics with tensorboard")
                for key,value in logs.items():
                    self._tb_writer.add_scalar(f"train_{key}_{tag}",value, steps)
                self._tb_writer.add_scalar(f"train_lr_{tag}", self.model.optimizer.lr.numpy(), steps)
                self._tb_writer.add_scalar(f"val_acc_{tag}", acc, steps)
                self._tb_writer.add_scalar(f"val_loss_{tag}", dev_loss, steps)


        if self.eval_on == "loss":
            if dev_loss <= self.best:
                self.wait = 0
                self.best = dev_loss
                self.save_ck(steps, tag, best_ck=True)
            else:
                self.wait += 1
        else:
            if acc >= self.best:
                self.wait = 0
                self.best = acc
                self.save_ck(steps, tag, best_ck=True)
            else:
                self.wait += 1
        logger.info(f"early stop count: {self.wait}/{self.patience}")
        logger.info(f"{tag}={steps}, best_on_eval_since({self.eval_on}): {self.best}")
        self.save_ck(steps, tag)
        if self.wait >= self.patience:
            logger.info("run out of patience, early stop")
            if self.use_tb:
                self._tb_writer.close()
            sys.exit(0)

    def save_ck(self, steps, tag="epoch", best_ck=False):
        sorted_indices, index2path = get_existing_cks(self.args.output_path, best_ck=best_ck)
        if len(sorted_indices) >= self.args.keep_ck_num:
            logger.info(
                f"since there are already {len(sorted_indices)} checkpoints saved that will be more than keep_ck_num={self.args.keep_ck_num}")
            logger.info(f"remove the oldest one: {index2path[sorted_indices[0]]}")
            os.remove(index2path[sorted_indices[
                0]])  # remove the oldest checkpoint, i.e., the one with the lowest epoch number
        # write_args(self.args.output_path, self.args)
        if best_ck:
            logger.info(
                f'save best model weights to {os.path.join(self.args.output_path, f"best_ck_at_{tag}_{steps}.h5")}')
            self.model.save_weights(os.path.join(self.args.output_path, f"best_ck_at_{tag}_{steps}.h5"),
                                    overwrite=True)
        else:
            logger.info(
                f'save model weights to {os.path.join(self.args.output_path, f"ck_at_{tag}_{steps}.h5")}')
            self.model.save_weights(os.path.join(self.args.output_path, f"ck_at_{tag}_{steps}.h5"),
                                    overwrite=True)

    def on_epoch_end(self, epoch, logs=None):
        if self.log_steps == -1:
            if self.args.do_eval:
                self.evaluate(epoch, tag="epoch",logs=logs)
        if not self.args.do_eval:
            # if do not do evaluate, the checkpoint at the end of epoch needs to be saved
            self.save_ck(epoch, tag="epoch")

def get_evaluator(x_eval, y_eval, args):
    add_filehandler_for_logger(args.output_path, logger)
    if args.task == "single-label-cls":
        return ClsEvaluator(x_eval, y_eval, args)
    elif args.task == "t2t":
        # it uses customize training toop, we do not need a evaluator here
        pass
    else:
        pass


import json, os
from tqdm import tqdm
import numpy as np
import logging, pickle
from ttt.utils import add_filehandler_for_logger

logging.basicConfig(
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)
logger = logging.getLogger(__name__)


def read_seq_single_cls_examples(data_path):
    texts = []
    labels = []
    with open(data_path, "r") as f:
        for line in tqdm(f, desc=f"reading from {data_path}"):
            example = json.loads(line.strip())
            texts.append(example["text"])
            labels.append(example["label"])
    return texts, labels


def convert_seq_single_cls_examples(data_path, tokenizer, max_seq_length, label2id):
    texts, labels = read_seq_single_cls_examples(data_path)
    y = np.array([label2id[label] for label in labels])
    encoded_texts = tokenizer(texts, padding=True, truncation=True, return_tensors="np",
                              max_length=max_seq_length)
    return texts, encoded_texts, y


def prepare_seq_single_cls_inputs(tokenizer, args, load_train_num=-1, tokenize_batch_size=None):
    logger.info("reading train set")
    train_texts, train_labels = read_seq_single_cls_examples(os.path.join(args.data_path, "train.json"))

    if load_train_num > 0:
        assert load_train_num <= len(train_texts), f"there are {len(train_texts)} training examples"
        logger.info(f"loading only {load_train_num} training examples out of the totaling {len(train_texts)}")
        train_texts = train_texts[:load_train_num]
        train_labels = train_labels[:load_train_num]

    label2id = {}
    logger.info("building label2id from train examples")
    for i, label in enumerate(list(set(train_labels))):
        label2id[label] = i

    logger.info("converting labels to its ids for train examples")
    y_train = np.array([label2id[label] for label in train_labels])
    logger.info(f"encoding train examples (num={len(train_texts)})")
    logger.info(f"using tokenizer with padding = True and truncation = True and max_length = {args.max_seq_length}")
    logger.info("This may take a while")
    # todo -> tokenize_batch_size
    encoded_train = tokenizer(train_texts, padding=True, truncation=True, return_tensors="np",
                              max_length=args.max_seq_length)

    if "token_type_ids" not in encoded_train:
        # we need this for roberta tokenizer that does not return token_type_ids
        encoded_train["token_type_ids"] = np.zeros(encoded_train["input_ids"].shape, dtype=np.int32)

    x_train = [encoded_train["input_ids"], encoded_train["token_type_ids"], encoded_train["attention_mask"]]

    to_return = {"x_train": x_train, "y_train": y_train, "label2id": label2id}

    if os.path.isfile(os.path.join(args.data_path, "val.json")):
        logger.info(f"found validation set in {os.path.join(args.data_path, 'val.json')}")
        logger.info(f"encoding validation examples")
        val_texts, encoded_val, y_eval = convert_seq_single_cls_examples(os.path.join(args.data_path, 'val.json'),
                                                                         tokenizer, args.max_seq_length, label2id)
        if "token_type_ids" not in encoded_val:
            # we need this for roberta tokenizer that does not return token_type_ids
            encoded_val["token_type_ids"] = np.zeros(encoded_val["input_ids"].shape, dtype=np.int32)

        x_eval = [encoded_val["input_ids"], encoded_val["token_type_ids"], encoded_val["attention_mask"]]

        to_return.update({"x_eval": x_eval, "y_eval": y_eval, "eval_examples": val_texts})
    # if os.path.isfile(os.path.join(args.data_path, "test.json")):
    #     logger.info(f"found test set in {os.path.join(args.data_path, 'test.json')}")
    #     logger.info(f"encoding test examples")
    #     test_texts, encoded_test, y_test = convert_seq_single_cls_examples(os.path.join(args.data_path, 'test.json'),
    #                                                                        tokenizer, args.max_seq_length, label2id)
    #     x_test = [encoded_test["input_ids"], encoded_test["token_type_ids"], encoded_test["attention_mask"]]
    #     to_return.update({"x_test": x_test, "y_test": y_test, "test_examples": test_texts})
    return to_return


def read_t2t_examples(data_path, source_field_name="text", target_field_name="label", with_target=True):
    source_texts = []
    target_texts = []
    other_attributes = []
    with open(data_path, "r",encoding="utf-8") as f:
        for line in tqdm(f, desc=f"reading from {data_path}"):
            example = json.loads(line.strip())

            source_texts.append(example.pop(source_field_name))
            # the </s> will be added when calling tokenizer, automatically when its add_special_tokens=True
            if with_target:
                target_texts.append(example.pop(target_field_name))
            other_attributes.append(example)
    return source_texts, target_texts, other_attributes


def convert_t2t_examples(data_path, tokenizer, args, with_target=True, return_meta=False):
    source_texts, target_texts, other_attributes = read_t2t_examples(data_path,
                                                                     source_field_name=args.source_field_name,
                                                                     target_field_name=args.target_field_name,
                                                                     with_target=with_target)

    # for pre - training t5, no need to append the </s> to the source sequence
    encoded_source = tokenizer(source_texts, padding=True, truncation=True, return_tensors="np",
                               max_length=args.max_src_length, add_special_tokens=not args.is_pretrain)

    encoded_target = None
    if with_target:
        # for making predictions purpose
        encoded_target = tokenizer(target_texts, padding=True, truncation=True, return_tensors="np",
                                   max_length=args.max_tgt_length, add_special_tokens=True)
    if return_meta:
        return source_texts, encoded_source, encoded_target, other_attributes

    return source_texts, encoded_source, encoded_target


def replace_with_special_token(encoded, special_token_id, replace_token_id=0):
    # replace_token_id: padding id
    ids = encoded["input_ids"]
    attention_mask = encoded["attention_mask"]

    for i in range(ids.shape[0]):
        indices = np.where(ids[i, :] == replace_token_id)[0]
        if indices.size == 0:
            ids[i, -1] = special_token_id
        else:
            ids[i, indices[0]] = special_token_id
            attention_mask[i, indices[0]] = 1
    return ids, attention_mask


def shift_to_right(input_ids, decoder_start_token_id):
    shifted_input_ids = np.zeros(input_ids.shape, dtype=input_ids.dtype)
    shifted_input_ids[..., 1:] = input_ids[..., :-1]
    shifted_input_ids[..., 0] = decoder_start_token_id
    # shifted_input_ids[shifted_input_ids == 0] = -100
    return shifted_input_ids


def tokenize_with_progress_bar(tokenizer, args, text_list, batch_size=1000):
    assert batch_size > 0
    encoded_return = {"input_ids": [], "attention_mask": []}
    batch = []
    # actual_max_seq = =args.max_src_length
    # tokenization in batch specified by batch_size, a bit different padding here compared to the at-one-go way, the batch here always pad to max_src_length although the longest one is not that long
    for idx, each_text in tqdm(enumerate(text_list), desc="tokenizing by batch...", total=len(text_list)):
        if (idx + 1) % batch_size == 0:
            batch.append(each_text)
            encoded = tokenizer(batch, padding="max_length", truncation=True, max_length=args.max_src_length,
                                add_special_tokens=not args.is_pretrain)
            encoded_return["input_ids"].extend(encoded["input_ids"])
            encoded_return["attention_mask"].extend(encoded["attention_mask"])
            batch = []
        else:
            batch.append(each_text)

    if batch != []:
        encoded = tokenizer(batch, padding="max_length", truncation=True, max_length=args.max_src_length,
                            add_special_tokens=not args.is_pretrain)
        encoded_return["input_ids"].extend(encoded["input_ids"])
        encoded_return["attention_mask"].extend(encoded["attention_mask"])

    assert len(encoded_return["input_ids"]) == len(text_list)

    encoded_return["input_ids"] = np.array(encoded_return["input_ids"])

    assert len(encoded_return["attention_mask"]) == len(text_list)

    encoded_return["attention_mask"] = np.array(encoded_return["attention_mask"])

    from transformers import BatchEncoding
    return BatchEncoding(data=encoded_return)

def tokenizet2t(tokenizer, args, source_text, target_text, batch_size=None):
    # this method is similar to T5Tokenizer.prepare_seq2seq_batch, it is rewritten here to get more control and flexibility
    logger.info(f"encoding source examples (num={len(source_text)})")
    logger.info(f"using tokenizer with padding = True and truncation = True and max_src_length = {args.max_src_length}")

    if batch_size is None:
        # if batch_size is None. refers to tokenization at one go. In this case, no progress bar but it can save memory potentially
        # for pre - training t5, no need to append the </s> to the source sequence
        logger.info("The tokenization is conducted at the backend without progress bar")
        logger.info("This may take a while")
        encoded_source = tokenizer(source_text, padding=True, truncation=True, return_tensors="np",
                                   max_length=args.max_src_length - 1 if not args.is_pretrain else args.max_src_length,
                                   add_special_tokens=not args.is_pretrain)
    else:
        encoded_source = tokenize_with_progress_bar(tokenizer, args, source_text, batch_size=batch_size)

    logger.info(f"encoding target examples (num={len(target_text)})")
    logger.info(f"using tokenizer with padding = True and truncation = True and max_tgt_length = {args.max_tgt_length}")
    if batch_size is None:
        logger.info("The tokenization is conducted at the backend without progress bar")
        logger.info("This may take a while")
        encoded_target = tokenizer(target_text, padding=True, truncation=True, return_tensors="np",
                                   max_length=args.max_tgt_length, add_special_tokens=True)
    else:
        encoded_target = tokenize_with_progress_bar(tokenizer, args, target_text, batch_size=batch_size)
    return encoded_source, encoded_target


def prepare_t2t_inputs(tokenizer, args, load_train_num=-1, tokenize_batch_size=None):
    logger.info("reading train set")
    source_texts_train, target_texts_train, _ = read_t2t_examples(os.path.join(args.data_path, "train.json"),
                                                                  source_field_name=args.source_field_name,
                                                                  target_field_name=args.target_field_name)
    if load_train_num > 0:
        # assert load_train_num <= len(source_texts_train), f"there are {len(source_texts_train)} training examples"
        logger.info(f"loading only {load_train_num} training examples out of the totaling {len(source_texts_train)}")
        source_texts_train = source_texts_train[:load_train_num]
        target_texts_train = target_texts_train[:load_train_num]

    encoded_source_train, encoded_target_train = tokenizet2t(tokenizer, args, source_texts_train, target_texts_train,
                                                             batch_size=tokenize_batch_size)

    # special_token_id = tokenizer.eos_token_id
    decoder_start_token_id = tokenizer.pad_token_id

    train_source_input_ids, train_source_attention_mask = encoded_source_train["input_ids"], encoded_source_train[
        "attention_mask"]

    train_target_input_ids, train_target_attention_mask = encoded_target_train["input_ids"], encoded_target_train[
        "attention_mask"]
    # this is pytorch's cross entropy's ignore index. to  figure out this in tensorflow-2.0
    # -> update: this can be found in transformers.modeling_tf_utils:TFCausalLanguageModelingLoss (since transformers 3.1.0)
    # shift_to_right(train_target_input_ids, decoder_start_token_id), this not needed since transformers 3.1.0 (it automatically takes care of these by simply passing labels argument to the model, just like in pytorch)
    # x_train = [train_source_input_ids, train_source_attention_mask,
    #            shift_to_right(train_target_input_ids, decoder_start_token_id), train_target_attention_mask]
    x_train = {"source_input_ids":train_source_input_ids, "source_attention_mask":train_source_attention_mask,
               "shifted_target_input_ids":shift_to_right(train_target_input_ids, decoder_start_token_id), "target_attention_mask":train_target_attention_mask}
    train_target_input_ids[train_target_input_ids == 0] = -100
    to_return = {"x_train": x_train, "y_train": {"target_input_ids":train_target_input_ids}}

    if args.do_eval:
        assert os.path.isfile(
            os.path.join(args.data_path, "val.json")), "do_eval=True, and no validation data (val.json) is found"

    if os.path.isfile(os.path.join(args.data_path, "val.json")):
        logger.info(f"found validation set in {os.path.join(args.data_path, 'val.json')}")
        logger.info(f"encoding validation examples")
        source_texts, encoded_source, encoded_target = convert_t2t_examples(
            os.path.join(args.data_path, 'val.json'),
            tokenizer, args)
        # add "</s>" add the end of each sequence
        source_input_ids, source_attention_mask = encoded_source["input_ids"], encoded_source["attention_mask"]
        target_input_ids, target_attention_mask = encoded_target["input_ids"], encoded_target["attention_mask"]
        # target_input_ids[target_input_ids == 0] = -100
        # in t5 trainer, eval lm labels are not used for calculating loss but for generation comparison, os we do not apply -100 replacement here
        # x_eval = [source_input_ids, source_attention_mask, shift_to_right(target_input_ids, decoder_start_token_id),
        #           target_attention_mask]
        x_eval = {"source_input_ids": source_input_ids, "source_attention_mask": source_attention_mask,
                   "shifted_target_input_ids": shift_to_right(target_input_ids, decoder_start_token_id),
                   "target_attention_mask": target_attention_mask}
        to_return.update({"x_eval": x_eval, "y_eval": {"target_input_ids":target_input_ids}, "eval_examples": {"source_texts":source_texts}})

    # if os.path.isfile(os.path.join(args.data_path, "test.json")):
    #     logger.info(f"found test set in {os.path.join(args.data_path, 'test.json')}")
    #     logger.info(f"encoding test examples")
    #     source_texts, encoded_source, encoded_target = convert_t2t_examples(
    #         os.path.join(args.data_path, 'test.json'),
    #         tokenizer, args)
    #     # add "</s>" add the end of each sequence
    #     source_input_ids, source_attention_mask = encoded_source["input_ids"], encoded_source["attention_mask"]
    #     target_input_ids, target_attention_mask = encoded_target["input_ids"], encoded_target["attention_mask"]
    #     # target_input_ids[target_input_ids == 0] = -100
    #     # in t5 trainer, eval lm labels are not used for calculating loss but for generation comparison, os we do not apply -100 replacement here
    #     x_test = [source_input_ids, source_attention_mask, shift_to_right(target_input_ids, decoder_start_token_id),
    #               target_attention_mask]
    #     to_return.update({"x_test": x_test, "y_test": target_input_ids, "test_examples": source_texts})
    return to_return


def get_with_prepare_func(tokenizer, args, prepare_func, load_train_num=-1, check_cache=False,
                          tokenize_batch_size=None):
    '''
        :param tokenizer:
        :param args:
        :return:
    '''
    args.is_load_from_data_cache = False
    if check_cache:
        if load_train_num > 0:
            data_cache_path = os.path.join(args.data_path,
                                           f"{args.model_select.replace('/', '-')}-data-{load_train_num}.pkl")
        else:
            data_cache_path = os.path.join(args.data_path, f"{args.model_select.replace('/', '-')}-data.pkl")
        args.data_cache_path = data_cache_path
        if os.path.isfile(data_cache_path):
            args.is_load_from_data_cache = True
            with open(data_cache_path, "rb") as f:
                logger.info(f"reading cached data from {data_cache_path}")
                logger.warning(
                    f"if you changed the max_seq_length/max_src_length/max_tgt_length, this may not correctly loaded, since the {data_cache_path} is pickled based on first time loading")
                to_return = pickle.load(f)
        else:
            to_return = prepare_func(tokenizer, args, load_train_num=load_train_num,
                                     tokenize_batch_size=tokenize_batch_size)
            with open(data_cache_path, "wb") as f:
                logger.info(f"caching data to {data_cache_path}")
                pickle.dump(to_return, f)
    else:
        to_return = prepare_func(tokenizer, args, load_train_num=load_train_num)
    return to_return

def get_inputs(tokenizer, args):
    add_filehandler_for_logger(args.output_path, logger)
    if args.task == "single-label-cls":
        inputs = get_with_prepare_func(tokenizer, args, prepare_seq_single_cls_inputs, check_cache=True,
                                       load_train_num=-1)
        args.input_seq_length = inputs["x_train"][0].shape[-1]
        args.label2id = inputs["label2id"]
        return inputs
    elif args.task == "t2t" or args.task == "translation" or args.task == "pretrain":
        args.is_pretrain = args.task == "pretrain"

        tokenize_batch_size=args.tokenize_batch_size if hasattr(args,"tokenize_batch_size") else None
        load_train_num=args.load_train_num if hasattr(args,"load_train_num") else -1

        data_dict = get_with_prepare_func(tokenizer, args, prepare_t2t_inputs, check_cache=True, load_train_num=load_train_num,
                                          tokenize_batch_size=tokenize_batch_size)
        args.source_sequence_length = data_dict["x_train"]["source_input_ids"].shape[-1]
        args.target_sequence_length = data_dict["y_train"]["target_input_ids"].shape[-1]
        return data_dict
    else:
        # when more tasks are supported -> todo
        pass



import argparse
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from nlp import load_dataset
import urllib.request
import os, json, sys
import shutil
import tarfile

# these have been tested and work fine. more can be added to this list to test
MODELS_SUPPORT = ["distilbert-base-cased", "bert-base-uncased", "bert-large-uncased",
                  "google/electra-base-discriminator",
                  "google/electra-large-discriminator", "albert-base-v2", "roberta-base",
                  "t5-small", "t5-base", "t5-large"]

# if using t5 models, the tasks has to be t2t* ones
TASKS_SUPPORT = ["single-label-cls", "t2t", "translation", "pretrain"]
# in the future, more schedulers will be added, such as warmupconstant, warmupcosine, etc.
LR_SCHEDULER_SUPPORT = ["warmuplinear", "warmupconstant", "constant", "constantlinear"]
# warmuplinear refers to increasing lr from 0 to a stage specified by warmup_ratio (ratio of total iterations) and then decreasing lr linearly
ADDITIONAL_DATA_SUPPORT = {"wmt_en_ro": "https://s3.amazonaws.com/datasets.huggingface.co/translation/wmt_en_ro.tar.gz"}


def ts2jsonl(data_path, t5_prefix_source=True):
    """
    t5_prefix_source: by default, t5_prefix_source, "translate English to Romanian: " is prepended to source sequence as in: https://arxiv.org/pdf/1910.10683.pdf
    """
    sets = ["train", "val", "test"]
    for set_name in sets:
        if os.path.isfile(os.path.join(data_path, set_name + ".source")):
            with open(os.path.join(data_path, set_name + ".source"), encoding="utf-8") as src:
                src_lines = src.readlines()
            with open(os.path.join(data_path, set_name + ".target"), encoding="utf-8") as src:
                tgt_lines = src.readlines()
            assert len(src_lines) == len(tgt_lines)
            if set_name == "train":
                #### for any train with examples more than 10000, we get 10000 to form a fixture training set (well suited for quick model development and prototyping)
                with open(os.path.join(data_path, set_name + "_fixture.json"), "w+", encoding="utf-8") as f:
                    for source, target in tqdm(zip(src_lines[:10000], tgt_lines[:10000])):
                        f.write(json.dumps({"source": "translate English to Romanian: " + source.strip(),
                                            "target": target.strip()}) + "\n")

            with open(os.path.join(data_path, set_name + ".json"), "w+", encoding="utf-8") as f:
                for source, target in tqdm(zip(src_lines, tgt_lines),
                                           desc=f"converting original data: {data_path} to jsonl formats"):
                    if t5_prefix_source:
                        source = "translate English to Romanian: " + source
                    f.write(json.dumps({"source": source.strip(), "target": target.strip()}) + "\n")


def data_check(args, sample_val_from_train=True, val_sample_portion=0.1,logger=None):
    if not os.path.isfile(os.path.join(args.data_path, "train.json")):
        # if it is from: https://huggingface.co/nlp/viewer, load it
        # if there is no validation set, by default, here a random sampling from the train (0.1 ratio) is used to form the val set
        # so far it works well for single-sequence cls datasets such as "glue/sst2", "20newsgroup", "ag_news", "imdb", "sentiment140", etc.
        # support more all kinds of other datasets that are availale in nlp, such as sequence-pair cls datasets (NLI), qa datasets, etc. -> todo
        set_name = args.data_path.split("/")[1:]
        try:
            dataset = load_dataset(*set_name)
            target_examples_dict = {}
            assert "train" in dataset, "not found train set in the given nlp dataset, make sure you give the correct name as listed in: https://huggingface.co/nlp/viewer/"
            label_names = dataset["train"].features["label"].names
            for set_key, examples in tqdm(dataset.items(),
                                          desc=f"not found the data locally, try to load {set_name} from nlp lib."):
                if set_key == "validation":
                    set_key = "val"
                if set_key not in target_examples_dict:
                    target_examples_dict[set_key] = []

                if set_key == "test" and "sst2" in set_name:
                    # sst2 does not have ground truths in test set from the nlp lib.
                    # here a special branch is processed here:
                    # to download a sst-test set with the same format as the train and val set loaded here
                    # from here: https://ucdcs-student.ucd.ie/~cwang/sst2-test.json
                    # for model testing
                    sst2_test = urllib.request.urlopen("https://ucdcs-student.ucd.ie/~cwang/sst2-test.json")
                    for line in sst2_test:
                        decoded_line = line.decode("utf-8").strip()
                        target_examples_dict[set_key].append(decoded_line)
                else:
                    for example in examples:
                        example["label"] = label_names[example["label"]]  # convert to raw label
                        if "sentence" in example:
                            example["text"] = example["sentence"]
                            del example["sentence"]
                        target_examples_dict[set_key].append(json.dumps(example))

            if sample_val_from_train and "val" not in target_examples_dict:
                train, val = train_test_split(target_examples_dict["train"], test_size=val_sample_portion,
                                              random_state=42)
                target_examples_dict["train"] = train
                target_examples_dict["val"] = val

            to_folder = os.path.join("data", *set_name)

            if not os.path.isdir(to_folder):
                os.makedirs(to_folder)

            splits_dict = {}
            for set_key, examples in tqdm(target_examples_dict.items(), desc="writing..."):
                with open(os.path.join(to_folder, set_key + ".json"), "w+") as target:
                    target.write("\n".join(examples))
                    splits_dict[set_key] = len(examples)
            with open(os.path.join(to_folder, "splits.txt"), "w+") as tgt:
                tgt.write(json.dumps(splits_dict, indent=2))

        except:
            if set_name[0] in ADDITIONAL_DATA_SUPPORT.keys():
                fstream = urllib.request.urlopen(ADDITIONAL_DATA_SUPPORT[set_name[0]])
                tfile = tarfile.open(fileobj=fstream, mode="r|gz")
                tfile.extractall(path="data")
                if set_name[0] == "wmt_en_ro":
                    ts2jsonl(args.data_path)
            else:
                if logger is not None:
                    logger.info("data not found")
                print("data not found")
                sys.exit(0)


def sanity_check(args,logger=None):
    # auto-construct some args
    # check if data exists
    data_check(args,logger=logger)

    output_folder = args.model_select + "_" + args.task + "_" + "-".join(args.data_path.split("/")[1:])
    output_path = os.path.join("tmp", output_folder)
    args.output_folder = output_folder
    args.output_path = output_path

    if args.do_train:
        if os.path.isdir(output_path):
            out = input(
                "Output directory ({}) already exists and is not empty, you wanna remove it before start training? (y/n)".format(
                    output_path))
            if out.lower() == "y":
                shutil.rmtree(output_path)
                os.makedirs(output_path, exist_ok=True)
            else:
                sys.exit(0)
        else:
            os.makedirs(output_path, exist_ok=True)

    assert args.model_select in MODELS_SUPPORT or os.path.isdir(args.model_select), F"set --model_select has to be in {MODELS_SUPPORT} or a local path where model's config and tokenizer_config exist"

    assert args.task in TASKS_SUPPORT, F"set --task to be in {TASKS_SUPPORT}"
    assert args.scheduler in LR_SCHEDULER_SUPPORT, F"set --scheduler to be in {TASKS_SUPPORT}"
    if "t5" in args.model_select:
        assert "t2t" in args.task or "translation" in args.task or "pretrain" in args.task, "t5 models (--model_select) only support t2t, translation, pretrain tasks (--task)"

    if "t5" not in args.model_select:
        assert "t2t" not in args.task, "BERT-like models (--model_select) only support non t2t tasks (--task)"

    if "translation" in args.task:
        assert "t5" in args.model_select, "translation task now is only compatible with T5 models"

    if "pretrain" in args.task:
        assert "t5" in args.model_select, "pretrain task now is only compatible with T5 models so far"


class Args:
    '''
    a Args class that maintain the same default args as argparse.ArgumentParser
    '''
    model_select = "bert-base-uncased"
    data_path = "data/glue/sst2"
    task = "single-label-cls"
    per_device_train_batch_size = 8
    eval_batch_size = 32
    num_epochs_train = 6
    log_steps = 400
    max_seq_length = 128
    max_src_length = 128
    max_tgt_length = 20
    source_field_name = "text"
    target_field_name = "label"
    lr = 5e-5
    warmup_ratio = 0.1
    patience = 20
    scheduler = "warmuplinear"
    seed = 122
    eval_on = "acc"
    keep_ck_num = 3
    ck_index_select = 0
    do_train = False
    do_eval = False
    do_test = False
    use_gpu = False
    use_tpu = False
    use_tb = False
    tpu_address = "x.x.x.x"

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def get_args():
    parser = argparse.ArgumentParser(description='Hyper params')

    parser.add_argument('--model_select', type=str, default="t5-small",
                        help='model select from MODEL_MAP')

    parser.add_argument('--data_path', type=str, default="data/glue/sst2",
                        help='data path')

    parser.add_argument('--task', type=str, default="t2t",
                        help='tasks available in TASKS_SUPPORT')

    parser.add_argument('--per_device_train_batch_size', type=int, default=8,
                        help='input batch size for training')

    parser.add_argument('--eval_batch_size', type=int, default=32,
                        help='input batch size for training')

    parser.add_argument('--num_epochs_train', type=int, default=6,
                        help='number of epochs to train')

    parser.add_argument('--log_steps', type=int, default=400,
                        help='logging steps for evaluation based on global step if it is not -1 and based on epoch if it is -1, and tracking metrics using tensorboard if use_tb is active')

    parser.add_argument('--max_seq_length', type=int, default=128,
                        help='maximum sequence length of samples in a batch for training')

    parser.add_argument('--max_src_length', type=int, default=128,
                        help='only working for t5-like t2t-based tasks, maximum source sequence length of samples in a batch for training')

    parser.add_argument('--max_tgt_length', type=int, default=20,
                        help='only working for t5-like t2t-based tasks, maximum target sequence length of samples in a batch for training')

    parser.add_argument('--source_field_name', type=str, default="text",
                        help='only working for t5-like t2t-based tasks, the source field name of the provided jsonl-formatted data')

    parser.add_argument('--target_field_name', type=str, default="label",
                        help='only working for t5-like t2t-based tasks, the target field name of the provided jsonl-formatted data')

    parser.add_argument('--lr', type=float, default=0.001,
                        help='default learning rate for fine-tuning as described in the T5 paper')

    parser.add_argument('--warmup_ratio', type=float, default=0.1,
                        help='warmup_ratio, only working if scheduler is not constant')

    parser.add_argument('--patience', type=int, default=20,
                        help='patience based on the log_steps')

    parser.add_argument('--scheduler', type=str, default="constant",
                        help='scheduler, default is constant as described in the T5 paper')

    parser.add_argument('--seed', type=int, default=122,
                        help='random seed')

    parser.add_argument('--eval_on', type=str, default="acc",
                        help='eval on for best ck saving and patience-based training early stop')

    parser.add_argument('--keep_ck_num', type=int, default=3,
                        help='keep_ck_num except for the best ck (evaluated on validation set using the metric specified by --eval_on')

    parser.add_argument('--ck_index_select', type=int, default=0,
                        help='ck_index_select, use the best one by default, negative one to specify a latest one, working when --do_test is active')

    parser.add_argument(
        "--do_train", action="store_true", help="Do training"
    )
    parser.add_argument(
        "--do_eval", action="store_true", help="do evaluation on validation set for saving checkpoint"
    )
    parser.add_argument(
        "--do_test", action="store_true", help="eval on test set if test set is availale"
    )
    parser.add_argument(
        "--use_gpu", action="store_true", help="use gpu?"
    )

    parser.add_argument(
        "--use_tpu", action="store_true", help="use tpu? "
    )
    parser.add_argument(
        "--use_tb", action="store_true", help="use tensorboard for tracking training process, default save to ./runs"
    )

    parser.add_argument('--tpu_address', type=str, default="x.x.x.x",
                        help='cloud tpu address if using tpu')

    parser.add_argument(
        "--default_store", action="store_true",
        help="Store datasets, weights, logs, and relevant details to folders by default?"
    )

    args = parser.parse_args()
    return args



from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name="pytriplet",
    version="0.0.3",
    author="Congcong Wang",
    author_email="wangcongcongcc@gmail.com",
    description="Fine-tuning Transformers with TPUs or GPUs acceleration, written in Tensorflow2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT License",
    url="https://github.com/wangcongcong123/ttt",
    download_url="https://github.com/wangcongcong123/ttt/releases/download/v0.0.3/pytriplet.tar.gz",
    packages=find_packages(),
    install_requires=[
        "tensorflow==2.3.0",
        "sklearn",
        "tqdm",
        "keras",
        "tensorboardX",
        "nlp",
        "sacrebleu",
        "transformers==3.1.0"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    keywords="Transformers, Tensorflow, TPUs acceleration"
)
# commands for uploading to pypi
# python setup.py sdist
# pip install twine
# twine upload dist/*



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from transformers import TFAutoModel, TFT5ForConditionalGeneration, T5Config
import os
from transformers.modeling_tf_utils import get_initializer


def get_lr_metric(optimizer):
    def lr(x, y):
        return optimizer.lr

    return lr


def create_sl_cls_model(model_name_or_path, input_seq_length, args):
    ## transformer encoder
    encoder = TFAutoModel.from_pretrained(model_name_or_path)

    encoder_config = encoder.config
    if not os.path.isfile(os.path.join(args.output_path, "config.json")):
        encoder_config.save_pretrained(args.output_path)

    input_ids = layers.Input(shape=(input_seq_length,), dtype=tf.int32)
    token_type_ids = layers.Input(shape=(input_seq_length,), dtype=tf.int32)
    attention_mask = layers.Input(shape=(input_seq_length,), dtype=tf.int32)

    if "distilbert" in args.model_select:
        # distilbert does not allow to pass token_type_ids
        sequence_outs = encoder(input_ids, attention_mask=attention_mask)[0]
    else:
        sequence_outs = encoder(input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)[0]

    # according to modeling_tf_bert:TFBertPooler. In transformers, models like ROBERTA and Electra do not ooffer direct outputs of pooled_output
    # to make it genelisable, the pooler is re-written here
    # this may not have a big effect on perf. if simply replacing the following pooler with this: pooled_output=sequence_outs[:, 0]
    pooled_output = tf.keras.layers.Dense(
        encoder_config.hidden_size,
        kernel_initializer=get_initializer(encoder_config.initializer_range),
        activation="tanh",
        name="dense",
    )(sequence_outs[:, 0])

    if hasattr(encoder_config, "hidden_dropout_prob"):
        pooled_output = tf.keras.layers.Dropout(encoder_config.hidden_dropout_prob)(pooled_output, training=True)
    else:
        pooled_output = tf.keras.layers.Dropout(encoder_config.dropout)(pooled_output, training=True)

    logits = tf.keras.layers.Dense(len(args.label2id), name="classifier", use_bias=False)(pooled_output)
    probs = layers.Activation(keras.activations.softmax)(logits)

    model = keras.Model(
        inputs=[input_ids, token_type_ids, attention_mask],
        outputs=probs,
    )

    loss = keras.losses.SparseCategoricalCrossentropy(from_logits=False)
    optimizer = keras.optimizers.Adam(lr=args.lr)
    model.compile(optimizer=optimizer, loss=loss, metrics=['accuracy', get_lr_metric(optimizer)])
    return model

def create_t2t_model(model_name_or_path, args, tokenizer=None,from_pretrained=True):
    ## transformer encoder
    if from_pretrained:
        encoder = TFT5ForConditionalGeneration.from_pretrained(model_name_or_path)
        encoder_config = encoder.config
    else:
        encoder_config = T5Config.from_pretrained(args.model_select)
        if tokenizer!=None:
            assert encoder_config.vocab_size==len(tokenizer)
            assert encoder_config.pad_token_id==tokenizer.pad_token_id
            assert encoder_config.eos_token_id==tokenizer.eos_token_id
            assert encoder_config.decoder_start_token_id==tokenizer.pad_token_id
        encoder = TFT5ForConditionalGeneration(encoder_config)
        # build the model with dummy_inputs
        encoder(encoder.dummy_inputs, training=False)

    if not os.path.isfile(os.path.join(args.output_path, "config.json")):
        encoder_config.save_pretrained(args.output_path)
    return encoder


def get_model(args,tokenizer=None,from_pretrained=True):
    if args.task == "single-label-cls":
        return create_sl_cls_model(args.model_select, args.input_seq_length, args)
    elif args.task == "t2t" or args.task == "translation" or args.task == "pretrain":
        return create_t2t_model(args.model_select, args, tokenizer=tokenizer,from_pretrained=from_pretrained)
    else:
        # when more task are supported -> todo
        pass



''''
this is a customize trainer for T5-like mode training,
in this class, the training loop is customized for more flexibility and control over
'''
import logging
import math
import os
import sys
import tensorflow as tf
from tqdm import tqdm
from sklearn.metrics import accuracy_score, classification_report
import numpy as np
from keras import backend as K

logging.basicConfig(
    format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.INFO
)
logger = logging.getLogger(__name__)

from ttt.utils import add_filehandler_for_logger, get_existing_cks

from tensorboardX import SummaryWriter
# for translation evaluation from: https://github.com/mjpost/sacrebleu
# which is also used in the original T5 paper
import sacrebleu

class T2TTrainer():
    def __init__(self, args):
        self.eval_on = args.eval_on
        assert self.eval_on in ["acc",
                                "bleu"], "now t2t training only supports --eval_on acc, bleu, only works when --do_eval=True"
        self.best = -np.Inf
        self.patience = args.patience
        self.wait = 0
        self.args = args
        self.use_tb = self.args.use_tb
        if self.use_tb:
            self._tb_writer = SummaryWriter(log_dir=os.path.join("runs", args.output_folder))
        self.scheduler = args.scheduler
        self.lr_to_reach = args.lr
        self.warmup_ratio = args.warmup_ratio
        add_filehandler_for_logger(args.output_path, logger)

    def save_ck(self, model, steps, tag="epoch", best_ck=False):
        sorted_indices, index2path = get_existing_cks(self.args.output_path, best_ck=best_ck)

        if len(sorted_indices) >= self.args.keep_ck_num:
            logger.info(
                f"there are already {len(sorted_indices)} checkpoints saved that will be more than keep_ck_num={self.args.keep_ck_num}")
            logger.info(f"hence, remove the oldest one: {index2path[sorted_indices[0]]}")
            os.remove(index2path[sorted_indices[
                0]])  # remove the oldest checkpoint, i.e., the one with the lowest epoch number

        if best_ck:
            logger.info(
                f'save best model weights to {os.path.join(self.args.output_path, f"best_ck_at_{tag}_{steps}.h5")}')
            model.save_weights(os.path.join(self.args.output_path, f"best_ck_at_{tag}_{steps}.h5"),
                               overwrite=True)
        else:
            logger.info(
                f'save model weights to {os.path.join(self.args.output_path, f"ck_at_{tag}_{steps}.h5")}')
            model.save_weights(os.path.join(self.args.output_path, f"ck_at_{tag}_{steps}.h5"),
                               overwrite=True)

    def train(self, model, strategy, tokenizer, inputs):
        x_train, y_train = inputs["x_train"], inputs["y_train"]
        num_train_examples = len(inputs["y_train"]["target_input_ids"])
        train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))

        if self.args.do_eval:
            assert "x_eval" in inputs and "y_eval" in inputs, "do_eval=True, and no validation data is found"
            x_val, y_val = inputs["x_eval"], inputs["y_eval"]
            val_dataset = tf.data.Dataset.from_tensor_slices((x_val, y_val))
            val_dataset = val_dataset.batch(self.args.eval_batch_size * strategy.num_replicas_in_sync)
            val_length = math.ceil(
                len(inputs["y_eval"]["target_input_ids"]) / (self.args.eval_batch_size * strategy.num_replicas_in_sync))

        global_batch_size = self.args.per_device_train_batch_size * strategy.num_replicas_in_sync
        train_dataset = train_dataset.shuffle(buffer_size=1024).batch(global_batch_size)
        train_dist_dataset = strategy.experimental_distribute_dataset(train_dataset)

        # THERE WILL BE exceptions when switching to distributed_dataset when running on tpus if
        # val_dist_dataset = strategy.experimental_distribute_dataset(val_dataset)
        train_length = math.ceil(num_train_examples / global_batch_size)
        self.steps_per_epoch = train_length
        # these are used for non-constant lr scheduler
        self.total_steps = self.steps_per_epoch * self.args.num_epochs_train
        self.warmup_steps = int(self.total_steps * self.warmup_ratio)

        with strategy.scope():
            optimizer = tf.keras.optimizers.Adam(lr=self.args.lr if self.scheduler.startswith("constant") else 0.0)
            def train_step(x_train,y_train):
                with tf.GradientTape() as tape:
                    outputs = model(inputs=x_train["source_input_ids"], attention_mask=x_train["source_attention_mask"],
                                    decoder_attention_mask=x_train["target_attention_mask"],
                                    labels=y_train["target_input_ids"], training=True, return_dict=True)
                    loss, logits = outputs.loss, outputs.logits  # return_dict=True
                    # bby default, no reduce operations (None) is applied inside the model
                    loss = tf.reduce_mean(loss, 0)

                gradients = tape.gradient(loss, model.trainable_variables)
                optimizer.apply_gradients(zip(gradients, model.trainable_variables))
                return loss

            @tf.function
            def distributed_train_step(x_train,y_train):
                per_replica_losses = strategy.experimental_run_v2(train_step, args=(x_train,y_train,))
                return strategy.reduce(tf.distribute.ReduceOp.SUM, per_replica_losses, axis=None)

            # evaluate
            def evaluate(steps, tag="epoch"):
                assert tag in ["epoch", "global_step"]
                gts = []
                preds = []
                for x_eval,y_eval in tqdm(val_dataset, total=val_length, desc="evaluating..."):
                    predictions = model.generate(input_ids=x_eval["source_input_ids"],
                                                 attention_mask=x_eval["source_attention_mask"],
                                                 max_length=self.args.max_tgt_length)
                    pred = [tokenizer.decode(ids) for ids in predictions]
                    gt = [tokenizer.decode(ids) for ids in y_eval["target_input_ids"]]
                    # labels (not -100 replaced since it is not used to calculate loss here)
                    preds.extend(pred)
                    gts.extend(gt)

                if self.eval_on == "bleu":
                    # bleu = 0
                    bleu = sacrebleu.corpus_bleu(preds, [gts])
                    eval_score = bleu.score
                else:
                    eval_score = accuracy_score(gts, preds)
                    logger.info(f"val_cls_report: {classification_report(gts, preds, digits=4)}")

                if self.use_tb:
                    self._tb_writer.add_scalar(f"val_{self.eval_on}_{tag}", eval_score, steps)

                logger.info("\n")
                logger.info(f"*******eval at {tag} = {steps} on validation dataset*********")
                logger.info(f"val_{self.eval_on}: {eval_score}")

                if self.eval_on == "acc" or self.eval_on == "bleu":
                    if eval_score >= self.best:
                        self.wait = 0
                        self.best = eval_score
                        logger.info(f"so far the best check point at {tag}={steps} based on eval_on {self.eval_on}")
                        self.save_ck(model, steps, tag, best_ck=True)
                    else:
                        self.wait += 1
                else:
                    raise ValueError("not support yet")

                logger.info(f"best so far({self.eval_on}): {self.best}")
                logger.info(f"early stop count: {self.wait}/{self.patience}")
                self.save_ck(model, steps, tag)
                if self.wait >= self.patience:
                    logger.info("run out of patience, early stop")
                    if self.use_tb:
                        self._tb_writer.close()
                    sys.exit(0)

            def update_lr(global_step):
                # already tested on tpu, works fine
                # global_step is dynamically passed here
                if global_step <= self.warmup_steps:
                    if self.scheduler == "warmuplinear" or self.scheduler == "warmupcostant":
                        inc = self.lr_to_reach / self.warmup_steps
                        K.set_value(optimizer.learning_rate, K.eval(optimizer.lr) + inc)
                else:
                    if self.scheduler == "warmuplinear" or self.scheduler == "constantlinear":
                        dec = self.lr_to_reach / (self.total_steps - self.warmup_steps)
                        K.set_value(optimizer.learning_rate, K.eval(optimizer.lr) - dec)
                # for "constant" scheduler, nothing to do here

            global_step = 0
            for epoch in tqdm(range(self.args.num_epochs_train), desc="epochs"):

                logger.info(f"start training at epoch = {epoch}")
                logger.info(f"global train batch size = {global_batch_size}")
                logger.info(f"using learning rate scheduler: {self.scheduler}")
                logger.info(
                    f"num_train_examples: {num_train_examples}, total_steps: {self.total_steps}, steps_per_epoch: {self.steps_per_epoch}")
                if self.scheduler != "constant":
                    logger.info(f"warmup_steps:{self.warmup_steps}")

                epoch_total_loss = 0.0
                num_batches = 0
                pbar = tqdm(enumerate(train_dist_dataset), total=train_length)
                for step, (x_train,y_train) in pbar:
                    # learning rate scheduler
                    update_lr(global_step)

                    loss = distributed_train_step(x_train,y_train)
                    epoch_total_loss += loss.numpy()
                    global_step += 1
                    num_batches += 1
                    pbar.set_description(
                        f"training - epoch {epoch + 1}/{self.args.num_epochs_train} iter {step}: train loss {loss.numpy():.5f}. lr {optimizer.lr.numpy():e}")

                    if self.args.log_steps != -1 and global_step % self.args.log_steps == 0:
                        if self.use_tb:
                            self._tb_writer.add_scalar("train_loss_global_step", epoch_total_loss / num_batches,
                                                       global_step)
                            self._tb_writer.add_scalar("train_lr_global_step", optimizer.lr.numpy(), global_step)

                        if self.args.do_eval:
                            evaluate(global_step, tag="global_step")
                        logger.info(f"train loss at global_step {global_step}: {epoch_total_loss / num_batches}")


                if self.args.log_steps == -1:
                    if self.args.do_eval:
                        evaluate(epoch, tag="epoch")
                    if self.use_tb:
                        self._tb_writer.add_scalar("train_loss_epoch", epoch_total_loss / num_batches,
                                                   global_step)
                        self._tb_writer.add_scalar("train_lr_epoch", optimizer.lr.numpy(), global_step)
                    logger.info(f"train loss at end of epoch {epoch}: {epoch_total_loss / num_batches}")

                if not self.args.do_eval:
                    # if do not do evaluate, the checkpoint at the end of epoch needs to be saved
                    self.save_ck(model, epoch+1, tag="epoch")
            if self.use_tb:
                self._tb_writer.close()
